/* Compile with:
 *   cc -o configtxt configtxt.c -Wall -Werror
 *
 * Run on all files:
 *   for i in c*[ip7]; do configtxt "$i" > "$i".txt; done
 *
 * This program attempts to decode the HP/Agilent 16510A logic
 * analyzer configuration files included with the inverse assemblers
 * for various processors.  The output is HP/Agilent-friendly text so
 * you can read it and type it into your logic analyzer's pod format
 * and symbol config.
 *
 * You will also need to load the companion inverse assembler file,
 * iXXXX_p or iXXXX_i to enable more than just the symbol decoding.
 *
 * This program has been tested on both the cXXXX_p and cXXXX_i
 * configuration files for general purpose probes and pre-processor
 * configurations.  The c6800_p configuration has been verified on a
 * real 6802.  All other configuration files are unverified, but they
 * do decode to text without any errors.  Spot checks with processor
 * pinouts and datasheets make sense.
 *
 * The purpose of all this is to allow use of the inverse assembler on
 * non-16510 cards for at least general purpose probing configuration.
 * This is in the cXXXX_p files.  The actual processor signal
 * assignments to specific pins can usually be inferred by the label
 * names and, where provided, the label symbols.
 *
 * The cXXXX_i files are for use with an HP/Agilent pre-processor that
 * has active electronics in it (mostly to create the STAT states),
 * but the contents may also provide additional clues for general
 * purpose probe users.  Anyone wanting to build an actual
 * pre-processor may also find them useful.  (Some of them don't look
 * too complicated.)
 *
 * It should be obvious, but you will want to use state mode on the
 * analyzer to capture ADDR, DATA, STAT, and any other desired
 * signals.  This program does not extract the clock configuration
 * from the raw config file, but hopefully that should be clear after
 * examining the available signals on the processor datasheet.
 *
 * For example, on a 6802 the falling edge of E is used as the clock
 * for the c6800_p configuration.  Pin assignments for STAT are:
 *
 *   STAT[2] - BA (bus available)
 *   STAT[1] - VMA (valid memory address)
 *   STAT[0] - R/W (read/not write)
 *
 *
 * General caveats:
 *
 * Most of the field locations in the raw config file were determined
 * by examining hex dumps and comparing to a few screen captures from
 * an actual 16510.  It's therefore likely this program will only
 * operate on the inverse assembler configuration files.  It has not
 * been tested on any other configuration files with more varied data.
 *
 * Some labels in the configuration file are disabled (maybe deleted?)
 * by whoever created them.  It could have been the original
 * developers.  Disabled labels and associated data appear to be
 * useful, so they are displayed anyway and flagged as disabled.
 *
 * A few sanity checks are performed as the file is parsed, but
 * anything unexpected could easily make memory accesses and data
 * interpretation go off the rails.  Memory bounds are not checked.
 *
 * In other words, this is a one-use hack.  The program has already
 * been run on all the configuration files with the output in .txt
 * files, so compiling is not needed.  This source is being provided
 * if anyone wants to make further enhancements.
 *
 *
 * v1.0  4/30/2018  MarkL@employees.org
 */


#include <unistd.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <endian.h>

void dump_hex(uint8_t *mem, int start_addr, int bytes, int addr_format)
/*
 * Utility - Dump a block of memory in hex.  Used during development.
 *
 * mem         is the block of memory
 * start_addr  is the starting label
 * bytes       is the length
 * addr_format is the length of the address field
 */
{
    int i = 0;
    int b;

    for ( ; ; ) {

	if (addr_format != 0) {
	    printf("%0*x: ", addr_format, start_addr + i);
	}

	for (b = 0; b < 16; b++) {
	    if (i+b < bytes)
		printf("%02x ", mem[i+b]);
	    else
		printf("   ");
	    if (b == 7)
		printf(" ");
	}

	printf(" ");

	for (b = 0; b < 16; b++)
	    if (i+b < bytes) {
		if (isprint(mem[i+b]))
		    putchar(mem[i+b]);
		else
		    putchar('.');
		if (b == 7)
		    printf(" ");
	    }

	putchar('\n');
    
	i += 16;

	if (i >= bytes) return;
    }
}


/* Max length of a config file.  Not checked.  Don't exceed.
 */
#define MAXBYTES	200000
uint8_t cfg[MAXBYTES];


struct pattern_s {
    uint8_t type;	/* format of a pattern is 2 */
    char name[16];
    uint8_t unk1;
    uint8_t is01[4];
    uint8_t dont_care[4];
};


struct label_s {
    uint8_t num;
    uint8_t unk1[1];
    char name[6];
    uint8_t unk2[7];
    uint8_t bitsdef_ptr;
    uint8_t unk3[4];
    uint8_t numbits;
    uint8_t enabled;
};


struct podbits_s {
    uint16_t a5_be;
    uint16_t a4_be;
    uint16_t a3_be;
    uint16_t a2_be;
    uint16_t a1_be;
};


void print_podbits(char *tag, uint16_t bits_be)
{
    uint16_t bits;
    int i;

    if (!bits_be) return;

    if (tag) printf("%s: ", tag);

    bits = be16toh(bits_be);
    
    for (i = 15; i >= 0; i--) {
	if (bits & 1 << i) {
	    printf("*");
	} else {
	    printf(".");
	}
	if (i == 8) printf(" ");
    }

    if (tag) printf("  ");
}


void print_labels(int pos)
{
    struct label_s *label;
    struct podbits_s *podbits;

    for ( ;; ) {

	label = (struct label_s *)&cfg[pos];

	if (label->numbits != 0) {

	    printf("Label %2d: %1.*s %c (%2d bits):  ", label->num,
		   (int)sizeof(label->name), label->name,
		   label->enabled ? ' ' : '?',
		   label->numbits);

	    //dump_hex(&cfg[0x93e +(label->bitsdef_ptr)], 0x93e +(label->bitsdef_ptr), 10, 0);

	    podbits = (struct podbits_s *)(&cfg[0x93e +(label->bitsdef_ptr)]);

	    print_podbits("A5", podbits->a5_be);
	    print_podbits("A4", podbits->a4_be);
	    print_podbits("A3", podbits->a3_be);
	    print_podbits("A2", podbits->a2_be);
	    print_podbits("A1", podbits->a1_be);
	    printf("\n");
	}
	    
	if (label->num == 0x13) break;

	pos += sizeof(struct label_s);
    }
}

struct label_s *find_label(int pos, char *find)
{
    struct label_s *label;

    for ( ;; ) {

	label = (struct label_s *)&cfg[pos];

	if (!strncmp(find, label->name, (int)sizeof(label->name))) {
	    return label;
	}
	    
	if (label->num == 0x13) break;

	pos += sizeof(struct label_s);
    }

    return NULL;
}


/*
 * Keep in mind the config file is big endian.
 */

int main(int argc, char **argv)
{
    FILE *f;

    int n;
    int end;
    int16_t sbytes;
    int pos;
    int need_newline;
    int config_checked;
    struct label_s *label;

    argv++;

    if (argc == 1) {
	f = stdin;
    } else {
	f = fopen(*argv, "r");
	if (!f) {
	    fprintf(stderr, "%s: %s\n", *argv, strerror(errno));
	    exit(1);
	}
    }

    /* First, skip the LIF header, 512 bytes.
     */
    n = fread(cfg, 512, 1, f);
    if (n != 1) {
	fprintf(stderr, "unexpected end of file\n");
	exit(1);
    }

    end = 0;
    config_checked = 0;

    /* Read the entire config file into the buffer.
     *
     * The config is split into a number of container records where
     * each record prefixed with a two-byte length.  Looks like 254
     * bytes is the max for each record and the last will have some
     * fraction of that.  EOF is indicated by all ones (FFFF or -1).
     *
     * We read one record at a time and strip the length field, so
     * we're left with just the pure config data in cfg[].
     */

    for ( ;; ) {

	/* First two bytes is the length.
	 */
	n = fread(&sbytes, 2, 1, f);
	if (n != 1) {
	    fprintf(stderr, "unexpected end of file\n");
	    exit(1);
	}
	
	/* Fix endian.
	 */
	sbytes = be16toh(sbytes);
	
	//printf("next sector contains %d bytes\n", sbytes);

	if (sbytes > 0xfe) {
	    fprintf(stderr, "warning: record length > 0xfe, reading anyway...\n");
	}

	if (sbytes == -1) {
	    //printf("EOF\n");
	    break;
	}

	n = fread(&cfg[end], sbytes, 1, f);
	//printf("read %d bytes\n", n);

	end += sbytes;

	if (end >= 0x24 + 10 && !config_checked) {
	    /* First record has been read.  Sanity check for "CONFIG".
	     */
	    if (strncmp((char *)&cfg[0x24], "CONFIG    ", 10)) {
		printf("error: not a config file; keyword \"CONFIG\" not at expected location\n");
		exit(1);
	    }
	    config_checked = 1;
	}

    }	

    //printf("Config bytes read: %d\n", end);

    printf("Title: %1.*s\n", 32, &cfg[0x04]);

    printf("\n");
    printf("Pod assigments (? = disabled)\n");
    printf("-----------------------------\n");

    /* First set of labels.
     */
    print_labels(0x7a);

    /* Second set of labels.  Do these do anything?  I don't think so.
     * Ignore for now.
     */
    //print_labels(0x246);

    pos = 0x4680;
    

    if (strncmp((char *)&cfg[pos], "SYMBOLS ", 8)) {
	printf("error: keyword \"SYMBOLS\" not at expected location\n");
	exit(1);
    }

    pos += 16;
    
    label = NULL;
    need_newline = 0;
    printf("\n");
    printf("Symbols\n");
    printf("-------\n");

    for ( ;; ) {

	/* First byte is type: 1=label name, 2=symbols that go with
	 * that label, 0=end of list.
	 */
	if (cfg[pos] == 0x01) {

	    /* Read label name.
	     */
	    if (need_newline == 1) {
		printf("\n");
		need_newline = 0;
	    }

	    label = find_label(0x7a, (char *)&cfg[pos+1]);
	    if (!label) {
		printf("error: can't find label: %1.*s\n", 6, &cfg[pos+1]);
		exit(1);
	    }

	    printf("Label: %1.*s", 6, &cfg[pos+1]);
	    if (!label->enabled) {
		printf("  (disabled)");
	    }
	    printf("\n");

	    pos += 8;
	}

	else if (cfg[pos] == 0x02) {

	    /* Read symbols to go with the previous label.
	     */
	    struct pattern_s *pattern;
	    int i;

	    if (!label) {
		printf("error: pattern with no previous label\n");
		exit(1);
	    }

	    pattern = (struct pattern_s *)&cfg[pos];
	    printf("       %1.*s  ", (int)sizeof(pattern->name), pattern->name);

	    //dump_hex(pattern->is01, pos, 8, 0);

	    for (i = label->numbits - 1; i >= 0; i--) {
		if (pattern->dont_care[3 - (i >> 8)] & 1 << i) {
		    printf("X");
		}
		else if (pattern->is01[3 - (i >> 8)] & 1 << i) {
		    printf("1");
		}
		else {
		    printf("0");
		}
	    }
	    printf("\n");
	    need_newline = 1;

	    pos += 26;
	}

	else if (cfg[pos] == 0) {
	    break;
	}

	else {
	    printf("error: unhandled symbol type 0x%02x\n", cfg[pos]);
	    exit(1);
	}
    }


    //dump_hex(&cfg[pos], pos, 16, 4);

    return 0;
}
