/* This program attempts to extract the relocatable (.R file) embedded
 * in the inverse assembler package files.
 *
 * The program takes two arguments.  Arg1 is the invasm package file
 * (iXXX file).  Arg2 is the output .R file.
 *
 * Compile with:
 *   cc -o extractr extractr.c -Wall -Werror
 *
 * Run on all files:
 *   for i in i*[ip7]; do extractr $i $i.r > $i.info; done
 *
 * Included in the package file are the settings for the Invasm Field
 * Option.  They are not part of the .R file, however, which is why IA
 * format and IALDOWN ask the user to set these options.  The field
 * option information is printed to stdout (and preserved in the .info
 * file, if the above line is used).
 *
 * This program has been run on all the inverse assembler files, so
 * this source is being provided for reference.  So far, only
 * i6800_p.r has been fully checked.
 *
 * v1.0  5/16/2018  MarkL@employees.org
 */


#include <unistd.h>
#include <stdio.h>
#include <stdint.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <endian.h>

void dump_hex(uint8_t *mem, int start_addr, int bytes, int addr_format)
/*
 * Utility - Dump a block of memory in hex.  Used during development.
 *
 * mem         is the block of memory
 * start_addr  is the starting label
 * bytes       is the length
 * addr_format is the length of the address field
 */
{
    int i = 0;
    int b;

    for ( ; ; ) {

	if (addr_format != 0) {
	    printf("%0*x: ", addr_format, start_addr + i);
	}

	for (b = 0; b < 16; b++) {
	    if (i+b < bytes)
		printf("%02x ", mem[i+b]);
	    else
		printf("   ");
	    if (b == 7)
		printf(" ");
	}

	printf(" ");

	for (b = 0; b < 16; b++)
	    if (i+b < bytes) {
		if (isprint(mem[i+b]))
		    putchar(mem[i+b]);
		else
		    putchar('.');
		if (b == 7)
		    printf(" ");
	    }

	putchar('\n');
    
	i += 16;

	if (i >= bytes) return;
    }
}


/* Max length of an input file.  Not checked.  Don't exceed.
 */
#define MAXBYTES	200000
uint8_t buf[MAXBYTES];


/* This gets overlayed onto the beginning of buf[].
 */
struct hdr_s {
    uint8_t unk1[4];
    uint8_t desc[32];
    uint8_t invasm_fld_opt;
    uint8_t r_magic[2];
} *hdr;


int main(int argc, char **argv)
{
    FILE *infile;
    FILE *outfile;

    int n;
    int buf_end;
    int16_t sbytes;
    int sanity_checked;

    if (argc != 3) {
	fprintf(stderr, "usage: %s INFILE OUTFILE\n", *argv);
	exit(1);
    }

    argv++;

    infile = fopen(*argv, "r");
    if (!infile) {
	fprintf(stderr, "%s: %s\n", *argv, strerror(errno));
	exit(1);
    }

    argv++;

    outfile = fopen(*argv, "w");
    if (!outfile) {
	fprintf(stderr, "%s: %s\n", *argv, strerror(errno));
	exit(1);
    }

    /* First, skip the LIF header, 512 bytes.
     */
    n = fread(buf, 512, 1, infile);
    if (n != 1) {
	fprintf(stderr, "unexpected end of file\n");
	exit(1);
    }

    buf_end = 0;
    sanity_checked = 0;
    hdr = (struct hdr_s *)buf;  // probably should be a union

    /* Read the entire config file into the buffer.
     *
     * The config is split into a number of container records where
     * each record prefixed with a two-byte length.  Looks like 254
     * bytes is the max for each record and the last will have some
     * fraction of that.  EOF is indicated by all ones (FFFF or -1).
     *
     * We read one record at a time and strip the length field, so
     * we're left with just the pure config data in buf[].
     */

    for ( ;; ) {

	/* First two bytes is the length.
	 */
	n = fread(&sbytes, 2, 1, infile);
	if (n != 1) {
	    fprintf(stderr, "unexpected end of file\n");
	    exit(1);
	}
	
	/* Fix endian.
	 */
	sbytes = be16toh(sbytes);
	
	//printf("next sector contains %d bytes\n", sbytes);

	if (sbytes > 0xfe) {
	    //fprintf(stderr, "warning: record length > 0xfe, reading anyway...\n");
	}

	if (sbytes == -1 || sbytes > 254) {
	    /* It seems to always be true that the EOF is always > 254 bytes.
	     *
	     * It's also possible the last record is always != 254, but we
	     * don't know for sure.
	     */
	    long pos;

	    pos = ftell(infile);
	    //printf("EOF, position = %ld (0x%04lx)\n", pos, pos);

	    break;
	}

	n = fread(&buf[buf_end], sbytes, 1, infile);

	if (sbytes != 254) {
	    //printf("read %d bytes\n", sbytes);
	}

	buf_end += sbytes;

	//dump_hex(buf, 0, 32, 4);

	if (buf_end >= sizeof(*hdr) && !sanity_checked) {
	    /* Header record has been read.  Sanity check for 82 03
	     */
	    if (hdr->r_magic[0] != 0x82 ||
		hdr->r_magic[1] != 0x03) {
		fprintf(stderr, "error: not an IA file, magic 82 03 not found\n");
		exit(1);
	    }
	    sanity_checked = 1;
	}

    }	

    //printf("Config bytes read: %d\n", buf_end);

    printf("Description: %1.*s\n", (int)sizeof(hdr->desc), hdr->desc);

    printf("Invasm Field Option: ");
    switch (hdr->invasm_fld_opt) {
    case 0xff:
	printf("A = No \"Invasm\" Field\n");
	break;
    case 0x00:
	printf("B = \"Invasm\" Field with no pop-up\n");
	break;
    case 0x01:
	printf("C = \"Invasm\" Field with pop-up. 2 choices in pop-up\n");
	break;
    case 0x02:
	printf("D = \"Invasm\" Field with pop-up. 8 choices in pop-up\n");
	break;
    default:
	printf("Unknown Invasm field option: 0x%02x\n", hdr->invasm_fld_opt);
	break;
    }

    //dump_hex(buf, 0, 64, 4);

    /* Create the .r file starting with the magic.  The number of
     * bytes is the buffer length that was read minus the junk at the
     * front before the .R magic 82 03.  (Note hdr == buf from above.)
     */
    n = fwrite(hdr->r_magic, buf_end - (hdr->r_magic - buf), 1, outfile);

    if (n == 0) {
	fprintf(stderr, "write to output file failed\n");
	exit(1);
    }
    fclose(outfile);

    return 0;
}
